@@ -35,11 +35,11 @@ class ICEBERG_BUNDLE_EXPORT ParquetWriter : public Writer {
 
   Status Close() final;
 
-  Status Write(ArrowArray array) final;
+  Status Write(ArrowArray* array) final;
 
-  std::optional<Metrics> metrics() final;
+  Result<Metrics> metrics() final;
 
-  std::optional<int64_t> length() final;
+  Result<int64_t> length() final;
 
   std::vector<int64_t> split_offsets() final;
 
