@@ -26,18 +26,44 @@
 #include <vector>
 
 #include "iceberg/iceberg_export.h"
+#include "iceberg/result.h"
+#include "iceberg/util/formatter.h"  // IWYU pragma: keep
 
 namespace iceberg {
 
 /// \brief A namespace in a catalog.
 struct ICEBERG_EXPORT Namespace {
   std::vector<std::string> levels;
+
+  bool operator==(const Namespace& other) const { return levels == other.levels; }
+
+  std::string ToString() const;
 };
 
+ICEBERG_EXPORT inline std::string ToString(const Namespace& ns) { return ns.ToString(); }
+
 /// \brief Identifies a table in iceberg catalog.
 struct ICEBERG_EXPORT TableIdentifier {
   Namespace ns;
   std::string name;
+
+  bool operator==(const TableIdentifier& other) const {
+    return ns == other.ns && name == other.name;
+  }
+
+  /// \brief Validates the TableIdentifier.
+  Status Validate() const {
+    if (name.empty()) {
+      return Invalid("Invalid table identifier: missing table name");
+    }
+    return {};
+  }
+
+  std::string ToString() const;
 };
 
+ICEBERG_EXPORT inline std::string ToString(const TableIdentifier& ident) {
+  return ident.ToString();
+}
+
 }  // namespace iceberg
