@@ -36,6 +36,7 @@
 #include "iceberg/result.h"
 #include "iceberg/schema_field.h"
 #include "iceberg/util/formattable.h"
+#include "iceberg/util/lazy.h"
 
 namespace iceberg {
 
@@ -122,18 +123,22 @@ class ICEBERG_EXPORT StructType : public NestedType {
       std::string_view name, bool case_sensitive) const override;
   using NestedType::GetFieldByName;
 
+  std::unique_ptr<Schema> ToSchema() const;
+
  protected:
   bool Equals(const Type& other) const override;
-  // TODO(nullccxsy): Lazy initialization has concurrency issues, need to add proper
-  // synchronization mechanism
-  Status InitFieldById() const;
-  Status InitFieldByName() const;
-  Status InitFieldByLowerCaseName() const;
+
+  static Result<std::unordered_map<int32_t, SchemaFieldConstRef>> InitFieldById(
+      const StructType&);
+  static Result<std::unordered_map<std::string_view, SchemaFieldConstRef>>
+  InitFieldByName(const StructType&);
+  static Result<std::unordered_map<std::string, SchemaFieldConstRef>>
+  InitFieldByLowerCaseName(const StructType&);
 
   std::vector<SchemaField> fields_;
-  mutable std::unordered_map<int32_t, SchemaFieldConstRef> field_by_id_;
-  mutable std::unordered_map<std::string_view, SchemaFieldConstRef> field_by_name_;
-  mutable std::unordered_map<std::string, SchemaFieldConstRef> field_by_lowercase_name_;
+  Lazy<InitFieldById> field_by_id_;
+  Lazy<InitFieldByName> field_by_name_;
+  Lazy<InitFieldByLowerCaseName> field_by_lowercase_name_;
 };
 
 /// \brief A data type representing a list of values.
@@ -150,6 +155,7 @@ class ICEBERG_EXPORT ListType : public NestedType {
   ~ListType() override = default;
 
   TypeId type_id() const override;
+  const SchemaField& element() const;
   std::string ToString() const override;
 
   std::span<const SchemaField> fields() const override;
@@ -520,4 +526,13 @@ ICEBERG_EXPORT std::shared_ptr<MapType> map(SchemaField key, SchemaField value);
 
 /// @}
 
+/// \brief Get the lowercase string representation of a TypeId.
+///
+/// This returns the same lowercase string as used by Type::ToString() methods.
+/// For example: TypeId::kBoolean -> "boolean", TypeId::kInt -> "int", etc.
+///
+/// \param id The TypeId to convert to string
+/// \return A string_view containing the lowercase type name
+ICEBERG_EXPORT std::string_view ToString(TypeId id);
+
 }  // namespace iceberg
